import argparse
import os
import shutil
import socket
import sys
from datetime import datetime
from difflib import unified_diff
from pathlib import Path
from typing import Iterator

from yaml.scanner import ScannerError

from hamstercage import Manifest
from hamstercage.hamstercage_exception import HamstercageException
from hamstercage.manifest import Host, Tag, Entry, DirEntry, SymlinkEntry, FileEntry
from hamstercage.utils import chmod, mode_to_str, short_date, print_table, ListEntry


class Hamstercage:
    """
    The main progam. Parses command line arguments and invokes sub-commands.
    """

    target: Path
    manifest_file: Path
    files: list
    hostname: str
    repo: Path
    tags: list

    def __init__(self):
        self.files = []
        self.hostname = socket.gethostname()
        self.manifest = None
        self.manifest_file = Path("hamstercage.yaml")
        self.repo = Path(".")
        self.tags = []
        self.target = Path("/")

    def main(self):
        parser = argparse.ArgumentParser(
            prog="hamstercage", description="Manage the hamster cage."
        )
        parser.add_argument(
            "-d",
            "--directory",
            type=Path,
            default="/",
            help="base directory of target files",
        )
        parser.add_argument(
            "-f",
            "--file",
            type=Path,
            default="hamstercage.yaml",
            help="manifest file to use",
        )
        parser.add_argument(
            "-n", "--hostname", default=socket.gethostname(), help="name of this host"
        )
        parser.add_argument(
            "-r", "--repo", type=Path, default=".", help="directory of file repo"
        )
        parser.add_argument("-t", "--tag", type=str, help="tags to apply/save")
        parser.add_argument(
            "-v", "--verbose", action="count", default=0, help="verbose output"
        )
        parser.set_defaults(func=None)

        subparsers = parser.add_subparsers(help="sub-command help")

        subparser = subparsers.add_parser(
            "add", help="add one or more files to the manifest"
        )
        subparser.set_defaults(func=self.add)
        subparser.add_argument(
            "-f",
            "--force",
            action="count",
            default=0,
            help="overwrite existing entries",
        )
        subparser.add_argument("files", nargs="+", help="files to add")

        subparser = subparsers.add_parser(
            "apply", help="apply files from repo to target"
        )
        subparser.set_defaults(func=self.apply)
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser(
            "diff", help="print differences between target and repo"
        )
        subparser.set_defaults(func=self.diff)
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser("init", help="create a new manifest")
        subparser.set_defaults(func=self.init)

        subparser = subparsers.add_parser(
            "list", aliases=["ls"], help="list manifest entries"
        )
        subparser.set_defaults(func=self.list)
        subparser.add_argument(
            "-l", "--long", action="count", default=0, help="list format long"
        )
        subparser.add_argument(
            "-t",
            "--tabs",
            action="count",
            default=0,
            help="separate columns with tabs instead of spaces",
        )
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser(
            "remove",
            aliases=["del", "rm"],
            help="remove one or more files from the manifest",
        )
        subparser.set_defaults(func=self.remove)
        subparser.add_argument("files", nargs="+", help="files to remove")

        subparser = subparsers.add_parser("save", help="save target files to repo")
        subparser.set_defaults(func=self.save)
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser(
            "tag",
            help="manage tags in the manifest",
        )
        subparser.set_defaults(func=None)

        tagparsers = subparser.add_subparsers(help="tag command help")

        subparser = tagparsers.add_parser(
            "tag",
            help="manage tags in the manifest",
        )
        subparser.set_defaults(func=self.tag_add)
        subparser.add_argument("name", nargs="1", help="tag name")
        subparser.add_argument(
            "-d",
            "--description",
            type=str,
            default="",
            help="description for this tag",
        )

        args = parser.parse_args()

        if args.func:
            self.target = args.directory
            self.manifest_file = args.file
            self.hostname = args.hostname
            self.repo = args.repo
            if args.tag:
                self.tags = [args.tag]

            try:
                return args.func(args)
            except HamstercageException as e:
                print(f"{e}", file=sys.stderr)
                return e.exit_code
        parser.print_help()
        return 64

    def add(self, args):
        """
        Add one or more entries to the manifest.
        :param args:
        :return:
        """
        self._load_manifest()
        if len(args.files) == 0:
            raise HamstercageException(f"Need at least one file to add", 64)
        if len(self.tags) != 1:
            raise HamstercageException(
                f"Need to specify exactly one tag to add files to", 64
            )

        self._run_hook(self.tags[0], "add", "pre")
        for file in args.files:
            entry = self._add_or_update(
                file, self.tags[0], ignore_existing=args.force > 0
            )
        self._run_hook(self.tags[0], "add", "post")
        self.manifest.dump()
        return 0

    def apply(self, args):
        """
        Copy (all/matching) manifest entries from the repo to the target.
        :return:
        """
        self._load_manifest()
        self._run_hooks("apply", "pre")
        for t, e in self._entries():
            e.apply(self._path_repo_absolute(t, e), self._path_target(e))
        self._run_hooks("apply", "post")
        return 0

    def diff(self, args):
        self._load_manifest()
        has_diff = False
        self.files = args.files

        self._run_hooks("diff", "pre")
        for t, e in self._entries():
            repo = self._path_repo_absolute(t, e)
            target = self._path_target(e)
            if not repo.is_file():
                continue  # non-files don't have a file under tags
            if target.exists() and repo.exists():
                diff = list(self._diff(target, repo))
                sys.stdout.writelines(diff)
                if diff:
                    has_diff = True
            else:
                if self._mtime_or_missing(repo, "---"):
                    has_diff = True
                if self._mtime_or_missing(target, "+++"):
                    has_diff = True
        self._run_hooks("diff", "post")
        return 1 if has_diff else 0

    def init(self, args) -> int:
        """
        Create a manifest file. Throw an error if the file already exists.
        :return:
        """
        if self.manifest_file.exists():
            raise HamstercageException(
                f'manifest file "{self.manifest_file}" already exists', 1
            )
        manifest = Manifest(str(self.manifest_file))
        manifest.hosts[self.hostname] = Host(self.hostname)
        manifest.hosts[self.hostname].tags = ["all"]
        manifest.tags = {
            "all": Tag.from_dict(
                "all", {"description": "files that apply to all hosts", "entries": {}}
            )
        }
        manifest.dump()
        return 0

    def list(self, args, file=None):
        """
        Print a list of all manifest entries
        :return:
        """
        if file is None:
            out = sys.stdout
        self._load_manifest()
        self.files = args.files
        items = {}
        for t, e in self._entries():
            repo = self._path_repo_absolute(t, e)
            target = self._path_target(e)
            if target not in items:
                items[target] = ListEntry(e, repo, t)
        if args.long == 0:
            for path in sorted(items):
                print(path, file=file)
        else:
            lines = []
            for path in sorted(items):
                item = items[path]
                entry = item.entry
                name = str(path)
                stat = path.stat()
                status = " "
                if not path.exists():
                    status = "!"
                type = "-"
                size = "0"
                if isinstance(entry, FileEntry):
                    if list(self._diff(path, item.repo)):
                        status = "*"
                    size = str(stat.st_size)
                if isinstance(entry, DirEntry):
                    name = name + "/"
                    type = "d"
                if isinstance(entry, SymlinkEntry):
                    name = name + " -> " + entry.target
                    type = "l"
                    if os.readlink(path) != entry.target:
                        status = "*"
                lines.append(
                    [
                        status,
                        mode_to_str(type, entry.mode),
                        entry.owner,
                        entry.group,
                        size,
                        short_date(int(stat.st_mtime)),
                        item.tag,
                        name,
                    ]
                )
            print_table(lines, align=["<", "<", "<", "<", ">"], file=file)
        # widths = [0] * 8
        # align = ["<", "<", "<", "<", ">", "<", "<", "<"]
        # for line in lines:
        #     for i in range(0, len(line)):
        #         widths[i] = max(widths[i], len(line[i]))
        # fs = []
        # for i in range(0, len(widths) - 1):
        #     fs.append(f"{{{i}:{align[i]}{widths[i]}}}")
        #     # f = f + f"{{0:{widths[i]}.{widths[i]}s}}"
        # fs.append(f"{{{len(widths)-1}}}")
        # for line in lines:
        #     # print("{0:} {1:}".format(*line), file=out)
        #     print(" ".join(fs).format(*line), file=out)
        return 0

    def remove(self, args):
        """
        Remove one or more entries from the manifest.
        :param args:
        :return:
        """
        self._load_manifest()
        if len(args.files) == 0:
            raise HamstercageException(f"Need at least one file to remove", 64)
        if len(self.tags) != 1:
            raise HamstercageException(
                f"Need to specify exactly one tag to remove files from", 64
            )

        for file in args.files:
            if file in self.manifest.tags[self.tags[0]].entries:
                repo = self._path_repo_absolute(
                    self.tags[0], self.manifest.tags[self.tags[0]].entries[file]
                )
                del self.manifest.tags[self.tags[0]].entries[file]
                repo.unlink()
            else:
                print(
                    f"Unable to remove {file}: no such entry in tag {self.tags[0]}",
                    file=sys.stderr,
                )
                return 71
        self.manifest.dump()
        return 0

    def save(self, args):
        """
        Copy (all/matching) manifest entries from the target to the repo.
        :return:
        """
        self._load_manifest()
        for (target, tag) in self._tags_for_targets().items():
            self._add_or_update(target, tag, require_existing=True)
        return 0

    def tag_add(self, args):
        self._load_manifest()
        if args.name in self.manifest.tags:
            raise HamstercageException(f"A tag named f{args.name} already exists")
        self.manifest.tags[args.name] = Tag(args.name, args.description)
        self.manifest.dump()
        return 0

    def _add_or_update(
        self, path: str, tag: str, ignore_existing=False, require_existing=False
    ) -> Entry:
        """
        Add a new file to, or update its contents and attributes in the repo.

        By default, it is an error if the repo file exists, and no error if the repo file doesn't exist.

        :param path: the target path of the file to be added/updated
        :param tag: the tag to update
        :param ignore_existing: it is not an error if the file already exists in the repo (default False)
        :param require_existing: it is an error if the file doesn't exist in the repo (default False)
        :return: the updated entry
        """
        if tag not in self.manifest.tags:
            raise HamstercageException(f"no tag {tag} in manifest")
        entries = self.manifest.tags[tag].entries
        target_path = self.target / path
        entry = Entry.entry(path, target_path)
        if require_existing:
            if entry.path not in entries:
                raise HamstercageException(
                    f"Unable to update {path}: no entry in manifest for tag {tag}"
                )
        elif not ignore_existing:
            if entry.path in entries:
                raise HamstercageException(
                    f"Unable to add {path}: already added to tag {tag}"
                )
        entries[path] = entry
        if entry.has_repo():
            repo = self._path_repo_absolute(self.tags[0], entry)
            try:
                self._mkdir_repo(self._path_repo_relative(self.tags[0], entry).parent)
                shutil.copy2(entry.target, repo, follow_symlinks=False)
                shutil.chown(repo, self.manifest.owner, self.manifest.group)
                repo.chmod(self.manifest.file_mode)
            except FileNotFoundError as e:
                raise HamstercageException(f"Unable to add {path}to tag {tag}: {e}", e)
        return entry

    @staticmethod
    def _diff(target, repo) -> Iterator[str]:
        """
        Generate a unified diff between the target and the repo file.

        :param target: the target file path
        :param repo: the repo file path
        :return: An iterator of diff lines
        """
        with open(target) as f:
            t = f.readlines()
        with open(repo) as f:
            r = f.readlines()
        return unified_diff(
            r,
            t,
            fromfile=str(repo),
            fromfiledate=Hamstercage._mtime(repo),
            tofile=str(target),
            tofiledate=Hamstercage._mtime(target),
        )

    def _entries(self):
        """
        Generator that produces all entries, potentially filtered by the list of files given on the command line. Only
        returns entries for the first match for a path.
        :return:
        """
        paths = {}
        for t in self.tags:
            for p, e in self.manifest.tags[t].entries.items():
                if not self._files_match(e):
                    continue
                if p in paths:
                    continue
                paths[p] = True
                yield t, e

    @staticmethod
    def _exists_status(path) -> str:
        """
        Return a string representing whether the path exists or not.
        :param path: to check
        :return: a space if the file exists, an exclamation sign otherwise
        """
        return " " if path else "!"

    def _files_match(self, entry: Entry) -> bool:
        """
        Return if the entry matches the files given on the command line. If the list of files is empty, any entry will
        match.

        :param entry: to check
        :return: True if the list of files matches this entry
        """
        if len(self.files) == 0:
            return True
        for f in self.files:
            if Path(entry.path).match(self._normalize_path(f)):
                return True
        return False

    def _load_manifest(self) -> None:
        """
        Load the manifest from the configured path. If the manifest had been loaded previously, do nothing.
        :return: None
        """
        if self.manifest:
            return
        try:
            self.manifest = Manifest(str(self.manifest_file))
            self.manifest.load()
            if len(self.tags) == 0:
                if self.hostname in self.manifest.hosts:
                    self.tags = self.manifest.hosts[self.hostname].tags
                else:
                    print(
                        f"Warning: No hostname entry for {self.hostname}",
                        file=sys.stderr,
                    )
        except FileNotFoundError as e:
            raise HamstercageException(
                f'Unable to load manifest from "{self.manifest_file}": {e}', 71
            )
        except ScannerError as e:
            raise HamstercageException(
                f'Unable to load manifest from "{self.manifest_file}": {e}', 71
            )

    def _mkdir_repo(self, path: Path) -> None:
        """
        Create a directory in the repo, copying owner/group/mode from the manifest file
        :param path: of the new directory
        :return: None
        """
        if str(path) != ".":
            self._mkdir_repo(path.parent)
        p = self.repo.joinpath(path)
        if not p.exists():
            p.mkdir(mode=self.manifest.dir_mode)
            chmod(str(p), self.manifest.dir_mode)
            shutil.chown(str(p), self.manifest.owner, self.manifest.group)

    @staticmethod
    def _mtime(path: Path) -> str:
        """
        Return the modification time of path as a string.
        :param path: of file
        :return: time and date in ISO8601 format
        """
        return datetime.fromtimestamp(path.stat().st_mtime).isoformat()

    def _mtime_or_missing(self, path: Path, prefix: str):
        """
        Print the modification date of path, or "missing" if the file doesn't exist.
        :param path: of the file
        :param prefix: string prefix to print
        :return:
        """
        if path.exists():
            print(f"{prefix} {str(path)}\t{self._mtime(path)}")
            return False
        print(f"{prefix} {str(path)}\tmissing")
        return True

    def _normalize_path(self, path: str) -> str:
        """
        Returns the normalized path for the given path.
        :param path:
        :return:
        """
        path = str(path)
        if path.startswith(str(self.target)):
            path = path[len(str(self.target)) :]
        if path.startswith("/"):
            path = path[1:]
        if path.endswith("/"):
            path = path[0:-1]
        return path

    def _path_repo_absolute(self, tag: str, entry: Entry) -> Path:
        """
        Return the absolute path for the file of the entry.
        :param tag: the tag for this entry
        :param entry: of the file
        :return: path of file
        """
        return self.repo.joinpath(self._path_repo_relative(tag, entry))

    def _path_repo_relative(self, tag: str, entry: Entry):
        """
        Return the relative path for the file of the entry.
        :param tag: the tag for this entry
        :param entry: of the file
        :return: path of file
        """
        return Path("tags") / tag / entry.path

    def _path_target(self, entry: Entry) -> Path:
        """
        Return the path to the target file for the entry.
        :param entry: of the file
        :return: path of file
        """
        return self.target.joinpath(entry.path)

    def _run_hook(self, tagname: str, cmd: str, step: str):
        """
        Execute the defined hook (if any) for the given command and step. If the hook is defined, but cannot be
        executed successfully, a HamstercageException is thrown.
        :param tagname: where the hook is defined
        :param cmd: command that is being executed
        :param step: pre or post
        :return: 0 if successful, any other exit code on failure.
        """
        tag = self.manifest.tags[tagname]
        hook = tag.find_hook(cmd, step)
        if hook:
            return hook.call(self.manifest, cmd, step, tag)
        return 0

    def _run_hooks(self, cmd: str, step: str):
        """
        Execute the defined hook (if any) for the given command and step for all tags. If the hook is defined, but
        cannot be executed successfully, a HamstercageException is thrown.
        :param cmd: command that is being executed
        :param step: pre or post
        :return: 0 if successful, any other exit code on failure.
        """
        for t in self.tags:
            r = self._run_hook(t, cmd, step)
            if r != 0:
                return r
        return 0

    def _tags_for_targets(self) -> dict:
        """
        Builds a list of files that should be processed. Resolved duplicate entries consistently.
        :return:
        """
        files: dict = {}
        for t, entry in self._entries():
            if entry.path in files:
                continue
            files[entry.path] = t
        return files


def main():
    h = Hamstercage()
    sys.exit(h.main())


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hamstercage', 'hamstercage.tests']

package_data = \
{'': ['*'], 'hamstercage.tests': ['repo/tags/all/*', 'repo/tags/foo/*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'importlib-resources>=5.7.1,<6.0.0']

entry_points = \
{'console_scripts': ['hamstercage = hamstercage:__main__.main']}

setup_kwargs = {
    'name': 'hamstercage',
    'version': '0.0.12',
    'description': 'Pets not cattle. A straightforward way to manage configuration files.',
    'long_description': '# Hamstercage - Config Management for Pets, not Cattle\n\n## Overview\nIf you work professionally with many machines and config, you like will have heard "cattle, not pets" as the philosophy for managing machines, VMs, etc., and you\'re likely using Ansible, Puppet, SaltStack, or another configuration management system that allows you to express configuration as code. This approach works well when you have many targets that share many traits, you have a lab where you can test configuration changes, and you have full time staff to take care of it all.\n\nHowever, if you\'re running a handful of boxes or VPSes for a small organisation, or just for yourself and your friends and family, your workflow might actually look quite different: you make changes to the live configuration of your web server, for example, and after you\'re satisfied that everything is working, you might want to save the key bits of config somewhere safe, so you can refer back to it later. Setting up any of the heavy tools can be cumbersome, especially for making quick provisional changes: in the worstcase scenario, you modify a file in the source repo, commit it, then run the tool to apply it to your machine.\n\nHamstercage aims to make it easy to save and restore your config by using a Git repo, by editing the config files directly on the target machine, then saving the new config into the repository. In other words: pets, not cattle.\n\nHamstercage is geared towards managing config files as complete files. To keep things simple, there are no facilities to update individual lines in files, update system configuration settings through some API, or other more complex logic. Hamstercage can be used to manage shell script files or binaries for custom tools, however.\n\nTo allow one repository to be used for multiple targets, sets of files can be managed. Each set is called a tag. You can select the tags to use each time you run Hamstercage. The manifest also contains a list of hostnames and the tags to use for each. This makes it possible to run the same Hamstercage command on multiple hosts, and have files be applied to each according to their respective purpose. \n\n## Theory of Operation\n\nHamstercage keeps a list of files under management in a **manifest** file, by default `hamstercage.yaml`. The various Hamstercage commands can be used to add, update, and remove files from the manifest, but you can always make changes manually, for example to correct the owner or mode of a file.\n\nManaged files are kept not as a single list, but as multiple lists, one for each **tag**. A tag is simply a label for a set of files. You are free to choose tags to organize sets of files as you see fit, for example, by purpose, application, operating system, or any other metric.\n\nIn addition to the managed files, the manifest also contains a list of **hostnames**, with the tags that should be applied to each hostname. This allows you to run the same Hamstercage command across multiple hosts and have files applied according to the hosts\' respective tags.\n\nThe contents of files is kept alongside the manifest, using the path of the target. For example, a target file `/etc/bash_profile` would be stored in `repo/all/etc/bash_profile`.\n\nWhile Hamstercage manages files very efficiently, there are some configuration tasks that require manipulation of existing files, refreshing a set of files, or restarting a daemon. Hamstercage does not attempt to implement any such actions. **Hooks** allow you to supply your own custom logic for these kinds of actions. You can define hooks to run before or after a command is executed (for example before a `save` or after an `apply`). Hooks are defined on tags, so you can cater the commands to the set of files that are managed. Hooks can be individual shell commands, any executable including scripts, or Python programs that are executed inside Hamstercages Python process.\n\n## Installation\n\nHamstercage is a Python module. You can install it from this source repository, or from PyPi.\n\n\n```shell\npip install hamstercage\n```\n\n## Using Hamstercage\n\nAfter installation, Hamstercage is available on the path as `hamstercage`.\n\n### Creating a new Hamstercage repo\n\nHamstercage needs a manifest file in order to track files managed through it. You can create this file manually, or you can have Hamstercage create a template file for you:\n\n```shell\nhamstercage init\n```\n\nThis creates `hamstercage.yaml` in the current directory.\n\n### The Hamstercage Manifest\n\nAll information about file entries is stored in the manifest file `hamstercage.yaml`. This [YAML](https://yaml.org) file consists of these top-level entries:\n* `hosts`: each entry describes one host, in particular the set of tags that apply to it\n* `tags`: each entry contains a list of files managed for this tag\n\nThis is a very minimal example with one host, one tag and one file:\n```yaml\nhosts:\n  testing.example.com:\n    description: \'\'\n    tags:\n    - all\ntags:\n  all:\n    description: files that apply to all hosts\n    entries:\n      foo.txt:\n        group: staff\n        mode: 0o644\n        owner: stb\n        type: file\n```\n\n#### Hosts\n\nEach entry in this dict defines one host. Hosts are distinguished by their fully-qualified hostname. When running Hamstercage, you can override the hostname by specifying the `-h` option. Each host definition can have these fields:\n* `description`: allows you to add a description or note to this entry. Optional.\n* `tags`: a list of one or more tags to apply to this host.\n\n#### Tags\n\nEach entry in this dict defines one tag. Tag names can be freely chosen; it is recommended to stick to alpha-numeric identifiers though, to keep the names compatible with shell scripts, etc. Each entry has these fields:\n* `description`: allows you to add a description or note to this entry. Optional.\n* `entries`: a dict of files managed through this tag. See below for details.\n* `hooks`: a dict of hook scripts that will be run. See below for details.\n\n#### Tags: Entries\n\nThe entries field in a tag entry describes files, directories, and symbolic links that are managed through that tag. Each entry can have these fields:\n* `group`: the group name owning this file.\n* `mode`: the access mode for this file. You need to specify the mode as an octal number (`0o644`) or as a string in quotes (`"644"`). If you simply give a number, it will be interpreted as a base-10 integer, leading to unexpected mode bits\n* `owner`: the user name owning this file.\n* `target`: the path the link points to, only applicable to `type`=`link`\n* `type`: must be one of `dir`, `file`, or `link`.\n\nNote that Hamstercage only manages symbolic links (soft links). Hard links are not supported.\n\n#### Tags: Hooks\n\nThe `hooks` dict in a tag allows you to hook scripts into Hamstercages execution. The name of each entry defines when to run the script; the contents of the dict define what command to execute.\n\nThe name consists of two parts separated by a dash: the prefix `pre` or `post`, and the name of the Hamstercage command. For example, the hook defined by the name `pre-save` will be executed when you run `hamstercage save`, just before the contents of the files will be copied from the target system to the repository. A hook `post-apply` will be run right after all files have been updated by the command `hamstercage apply`.\n\nHamstercage searches for a matching hook definition in this order:\n1. for the exact match *step*`-`*command* (for example, `pre-save`)\n2. for the wildcard match `*-`*command* (for example, `*-save`)\n3. for the wildcard match *step*`-*` (for example, `pre-*`)\n4. for the fallback wildcard `*`.\n\nThe first match found will be executed.\n\nEach entry has these fields:\n* `command`: the command or script to execute.\n* `description`: allows you to add a description or note to this entry. Optional.\n* `type`: the type of command or script to run.\n  * `exec`: run the program specified by `command`. If `command` is not an absolute path, the system will search for it on the `PATH`. See [subprocess.Popen](https://docs.python.org/3/library/subprocess.html#subprocess.Popen) for details. The command will be invoked with the Hamstercage command, the step, and the tag the hook is defined in as parameters, for example `.../myhook.sh apply post all`.\n  * `python`: `command` should specify a Python script. A relative path is interpreted relative to the repository directory. See below for global variables available to the script.\n  * `shell`: run `command` through the shell. Typically, this will be the shell of the user running Hamstercage.\n\nThe `exec` and `shell` hooks receive the following environment variables:\n* `HAMSTERCAGE_CMD`: the command being executed\n* `HAMSTERCAGE_MANIFEST`: the manifest file path name\n* `HAMSTERCAGE_HOOK`: the name of the hook entry\n* `HAMSTERCAGE_REPO`: the repo directory, which is the directory the manifest file lives in\n* `HAMSTERCAGE_STEP`: the step (`pre` or `post`)\n* `HAMSTERCAGE_TAG`: the tag name this hook is defined in\n\nThe `python` script receives these global variables:\n* `cmd`: the command being executed\n* `manifest`: the manifest object\n* `hook`: the name of the hook entry\n* `repo`: the repo directory, which is the directory the manifest file lives in\n* `step`: the step (`pre` or `post`)\n* `tag`: the tag object this hook is defined in\n* `__file__`: the Python hook file\n* `__name__`: the constant `__hamstercage__`.\n\n\n\n## Command Reference\n\nHamstercage commands take a number of options and parameters.\n\n### General Options\n\n#### Target Directory `-d` / `--directory`\n\nThe directory where the managed files live, default `/`. You can set this to a different path when restoring config from a backup, or when managing a jail or container from the outside.\n\n#### Hostname `-n` / `--hostname`\n\nThe hostname of the local host. By default the value of `hostname(1)`. When applying files to the target directory, the hostname is used to obtain the default list of tags to use. Specifying this can be useful when restoring a host from backup, or when working on a container or jail.\n\n#### Manifest File `-f` / `--file`\n\nThe name of the manifest file that tracks all managed files, as well as settings for the repository. Default is `./hamstercage.yaml`.\n\n#### Repository Base Directory `-r` / `--rep`\n\nAll managed files are stored under this directory. The default is the same directory the manifest file lives in.\n\n#### Tags `-t` / `--tag`\n\nWhen adding files or applying them to the target directory, one or more tags can be specified. See the `add` and `apply` commands.\n\n### Verbose output `-v` / `--verbose`\n\nBy default, Hamstercage will only print warnings or error. With this flag, progress will be reported on all operations.\n\n### Commands\n\n#### Adding Files `add`\n\n```shell\nhamstercage -t mytag add file...\n```\n\nAdd one or more files to the repository. You need to specify exactly one tag, which determines the tag the file will be added to. The file will be added to the manifest, and the contents of the file copied to the repository.\n\n#### Applying Files `apply`\n\n```shell\nhamstercage apply [file...]\n```\n\nApply files from the repository to the target. Without any further options and parameters, will apply all files from the default tags for this hostname. Specify one or more tags to limit files to those in those tags, or specify filenames that should be applied.\n\nAdding a file that is already in the manifest for the given tag is an error.\n\n#### Determining Differences `diff`\n\n```shell\nhamstercage diff [file...]\n```\nPrints out a unified diff between all files in the repository and in the target. This allows you to preview what changed `apply` or `save` will make.\n\n#### Initialize a New Repository `init`\n\n```shell\nhamstercage init\n```\n\nCreates a new manifest file, by default `./hamstercage.yaml`. Combine this with `git init` to create a new Hamstercage repository.\n\n#### List Entries `list`\n\n```shell\nhamstercage list [-l]\n```\nLists all entries in the manifest. When `-l` is given, will print more details on each entry.\n\n#### Remove Files From Repository `remove`\n\n```shell\nhamstercage -t tag remove file...\n```\n\nRemoves one or more files from the specified tag. Does not change files in the target.\n\n#### Save Target to Repository `save`\n\n```shell\nhamstercage save [file...]\n```\n\nUpdates the manifest entries and repository files from the target.\n\n#### Adding Tags `tag add`\n\nAdds a new tag to the manifest.\n\n```shell\nhamstercage tag add [-d description] [file...]\n```\n\n# Developing Hamstercage\n\n## Poetry For Dependency Management and Building\n\nThe project uses [Poetry](https://python-poetry.org), which you should install locally. After installing Poetry, you can install all necessary dependencies:\n```shell\npoetry install\n```\n\n\n## Source Code Formatting With Black\n\nThe GitHub workflow checks source code formatting with [black](https://github.com/psf/black).\n\nTo format all code automatically:\n```shell\npoetry run black .\n```\n\nWhen working on the code, you might want to [configure your IDE to automatically reformat the code with black](https://black.readthedocs.io/en/stable/integrations/editors.html).',
    'author': 'Stefan Bethke',
    'author_email': 'stb@lassitu.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://hamstercage.io',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
